/**
 * Copyright (c) 2013-2015 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.extension.autoloader;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.List;
import java.util.TimerTask;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;
import org.ow2.petals.microkernel.api.communication.JMXService;
import org.ow2.petals.microkernel.api.jbi.management.DeploymentService;
import org.ow2.petals.microkernel.api.jbi.management.InstallationService;
import org.ow2.petals.microkernel.api.jbi.management.JbiArchiveType;
import org.ow2.petals.microkernel.api.jbi.management.PackageHelper;
import org.ow2.petals.microkernel.api.jbi.management.XMLResult;
import org.ow2.petals.microkernel.api.util.LoggingUtil;
import org.ow2.petals.microkernel.api.util.NioUtil;

/**
 * The {@link TimerTask} scanning the installation directory
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class InstallDirectoryScanningTask extends TimerTask {

    private final DeploymentService deploymentService;

    private final InstallationService installationService;

    private final JMXService jmxService;

    private final File installDir;

    private final ArchiveFileFilter archiveFilter = ArchiveFileFilter.getInstance();

    private final List<File> previousArchives = new ArrayList<File>();

    private final LoggingUtil logger;

    public InstallDirectoryScanningTask(final DeploymentService deploymentService,
            final InstallationService installationService, final JMXService jmxService,
            final File installDir, final LoggingUtil logger) {
        super();

        this.deploymentService = deploymentService;
        this.installationService = installationService;
        this.jmxService = jmxService;
        this.installDir = installDir;
        this.logger = logger;
    }

    @Override
    public void run() {
        final File[] newInstalls = this.installDir.listFiles(this.archiveFilter);
        if (newInstalls != null) {
            final List<File> newInstallList = new ArrayList<File>();

            synchronized (this.previousArchives) {

                // get only new archives
                for (final File archive : newInstalls) {
                    if (!this.previousArchives.contains(archive)) {
                        newInstallList.add(archive);
                    }
                }
                this.previousArchives.clear();

                // remove archives that are not completely copied in the install
                // directory
                final List<File> deleteInstallList = new ArrayList<File>();
                for (final File newArchive : newInstallList) {
                    if (!NioUtil.checkAccess(newArchive)) {
                        deleteInstallList.add(newArchive);
                    }
                }
                newInstallList.removeAll(deleteInstallList);

                if (!newInstallList.isEmpty()) {
                    this.install(newInstallList);
                    this.previousArchives.addAll(newInstallList);
                }
            }
        } else {
            this.logger.warning("Error when the autoloader scanned the directory '"
                    + this.installDir.getPath() + "'. Perhaps, the directory was deleted !");
        }
    }

    /**
     * Install the given list of files.
     * 
     * @param archiveFiles
     *            the list of JBI archives
     */
    private void install(List<File> archiveFiles) {
        this.logger.start();

        final List<File> slToInstall = this.getJBIArchives(JbiArchiveType.SL_TYPE, archiveFiles);
        for (final File archive : slToInstall) {
            try {
                this.performInstallSL(archive);
            } catch (final Throwable e) {
                this.logger.error("Failed to auto-install of a shared library", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.logger.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(slToInstall);

        final List<File> compToInstall = this.getJBIArchives(JbiArchiveType.COMPONENT_TYPE,
                archiveFiles);
        for (final File archive : compToInstall) {
            try {
                this.performInstallComponent(archive);
            } catch (final Throwable e) {
                this.logger.error("Error during the auto- installation of a component", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.logger.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(compToInstall);

        final List<File> saToInstall = this.getJBIArchives(JbiArchiveType.SA_TYPE, archiveFiles);
        for (final File archive : saToInstall) {
            try {
                this.performDeploySA(archive);
            } catch (final Throwable e) {
                this.logger.error("Error during the deployment of a service assembly: ", e);
            } finally {
                try {
                    archive.delete();
                } catch (final SecurityException e) {
                    this.logger.error(e.getMessage(), e);
                }
            }
        }
        archiveFiles.removeAll(saToInstall);

        // clean the remaining archives
        for (final File archive : archiveFiles) {
            try {
                this.logger.warning("The archive '" + archive.getName()
                        + "' remains in the install list. PEtALS deletes it");
                archive.delete();
            } catch (final SecurityException e) {
                this.logger.error(e.getMessage(), e);
            }
        }

        this.logger.end();
    }

    /**
     * Install the specified shared library in the JBI container Any exception
     * will cause the abortion of the process. A file "fileName.success" will be
     * created if success A file "fileName.failed" will be created otherwise,
     * containing the related exception
     * 
     * @param slFile
     *            The shared library
     */
    private void performInstallSL(File slFile) {
        this.logger.call();

        try {
            this.installationService.installSharedLibrary(slFile.toURI().toString());
        } catch (final Throwable e) {
            this.logger.error("Failed to auto-install a shared library");
        }
    }

    /**
     * Install the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param componentFile
     *            The component
     * @throws MalformedURLException
     * @throws ReflectionException
     * @throws MBeanException
     * @throws InstanceNotFoundException
     */
    private void performInstallComponent(File componentFile) throws MalformedURLException,
            InstanceNotFoundException, MBeanException, ReflectionException {
        this.logger.call();
        final ObjectName newInstallerName = this.installationService.loadNewInstaller(componentFile
                .toURI().toURL().toExternalForm());
        final Object[] nullObjects = new Object[0];
        final String[] nullStrings = new String[0];
        final MBeanServer jmxServer = this.jmxService.getLocalJMXServer();
        final Object result = jmxServer.invoke(newInstallerName, "install", nullObjects,
                nullStrings);
        if (result instanceof ObjectName) {
            final ObjectName componentName = (ObjectName) result;
            jmxServer.invoke(componentName, "start", nullObjects, nullStrings);
        } else {
            throw new RuntimeException(
                    "The install() invocation on the component installer has returned an unexpected object : "
                            + result);
        }
    }

    /**
     * Get the set of files to install according to the JBI artefact type.
     * 
     * @param jbiType
     *            The expected JBI artefact
     * @return The list of selected JBI artefact files
     * @throws IOException
     * @throws JBIDescriptorException
     * @throws MalformedURLException
     */
    private List<File> getJBIArchives(JbiArchiveType jbiType, List<File> archivesToUninstall) {
        final List<File> archives = new ArrayList<File>();

        for (final File archive : archivesToUninstall) {
            try {
                final Jbi descriptor = PackageHelper.loadDescriptor(archive.toURI().toURL());
                if (JbiArchiveType.SL_TYPE.equals(jbiType) && descriptor.getSharedLibrary() != null) {
                    archives.add(archive);
                } else if (JbiArchiveType.COMPONENT_TYPE.equals(jbiType)
                        && descriptor.getComponent() != null) {
                    archives.add(archive);
                } else if (JbiArchiveType.SA_TYPE.equals(jbiType)
                        && descriptor.getServiceAssembly() != null) {
                    archives.add(archive);
                }
            } catch (final PetalsException e) {
                this.logger.warning("Unable to read a JBI descriptor", e);
            } catch (final IOException e) {
                this.logger.warning("Unable to read a JBI descriptor", e);
            } catch (JBIDescriptorException e) {
                this.logger.warning("Unable to read a JBI descriptor", e);
            }
        }

        return archives;
    }

    /**
     * Deploy the specified Archive in the JBI container and start it . Any
     * exception will cause the abortion of the process. A file
     * "fileName.success" will be created if success A file "fileName.failed"
     * will be created otherwise, containing the related exception
     * 
     * @param saFile
     *            the archive file
     * @throws Exception
     */
    private void performDeploySA(File saFile) throws Exception {
        this.logger.call();

        final String xml = this.deploymentService.deploy(saFile.toURI().toString());

        if (xml.indexOf(XMLResult.TaskResult.FAILED.toString()) > -1) {
            throw new RuntimeException("Failed to auto-deploy a service assembly");
        } else {
            final String saName = xml.substring(xml.indexOf("<loc-param>") + 11,
                    xml.indexOf("</loc-param>"));
            this.deploymentService.start(saName);
        }
    }

}
